from typing import Generator, Iterable, Any
import logging, copy, sys

from mirp._data_import.generic_file import ImageFile
from mirp.settings.generic import SettingsClass
from mirp.utilities.parallel import (
    start_parallel_cluster, parse_parallel_backend, message_parallel_process, cluster_exists, shutdown_cluster)
from mirp.utilities.parallel_ray import ray_remote, ray_init, ray_is_initialized, ray_get, ray_shutdown
from mirp._workflows.standardWorkflow import StandardWorkflow


def extract_features(
        write_features: None | bool = None,
        export_features: None | bool = None,
        write_dir: None | str = None,
        num_cpus: None | int = None,
        **kwargs
) -> None | list[Any]:
    """
    Compute features from regions of interest in images. This function is a wrapper around
    :func:`~mirp.extractFeaturesAndImages.extract_features_and_images`.

    Parameters
    ----------
    write_features: bool, optional
        Determines whether features computed from images should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    export_features: bool, optional
        Determines whether features computed from images should be returned by the function.

    write_dir: str, optional
        Path to directory where feature tables should be written. If not set, feature tables are returned by this
        function. Required if ``write_features=True``.

    num_cpus: int, optional, default: None
        Number of CPU nodes that should be used for parallel processing. Image processing and feature computation can be
        parallelized using the ``ray`` package. If a ray cluster is defined by the user, this cluster will be used
        instead. By default, images are processed sequentially.

    **kwargs:
        Keyword arguments passed for importing images and masks (
        :func:`mirp.data_import.import_image_and_mask.import_image_and_mask`) and configuring settings:

        * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
        * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
        * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
        * image interpolation / resampling (
          :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
          :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
        * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
        * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)
        * feature computation / extraction (
          :class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`)

    Returns
    -------
    None | list[Any]
        List of feature tables, if ``export_features=True``.

    See Also
    --------
    :func:`~mirp.extractFeaturesAndImages.extract_features_and_images`

    """
    return extract_features_and_images(
        write_features=write_features,
        export_features=export_features,
        write_images=False,
        export_images=False,
        write_dir=write_dir,
        num_cpus=num_cpus,
        **kwargs
    )


def extract_features_generator(
        write_features: bool = False,
        export_features: bool = True,
        **kwargs
):
    """
    Compute features from regions of interest in images. This generator is a wrapper around
    :func:`~mirp.extractFeaturesAndImages.extract_features_and_images_generator`.

    Parameters
    ----------
    write_features: bool, default: False
        Determines whether features computed from images should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    export_features: bool, default: True
        Determines whether features computed from images should be returned by the function.

    **kwargs:
        Keyword arguments passed for importing images and masks (
        :func:`mirp.data_import.import_image_and_mask.import_image_and_mask`) and configuring settings:

        * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
        * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
        * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
        * image interpolation / resampling (
          :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
          :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
        * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
        * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)
        * feature computation / extraction (
          :class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`)

    Returns
    -------
    None | list[Any]
        List of feature tables, if ``export_features=True``.

    See Also
    --------
    :func:`~mirp.extractFeaturesAndImages.extract_features_and_images_generator`

    """
    yield from extract_features_and_images_generator(
        write_features=write_features,
        export_features=export_features,
        write_images=False,
        export_images=False,
        **kwargs
    )


def extract_images(
        write_images: None | bool = None,
        export_images: None | bool = None,
        write_dir: None | str = None,
        image_export_format: str = "dict",
        num_cpus: None | int = None,
        **kwargs
):
    """
    Process images and masks. This function is a wrapper around
    :func:`~mirp.extractFeaturesAndImages.extract_features_and_images`.

    Parameters
    ----------
    write_images: bool, optional
        Determines whether processed images and masks should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    export_images: bool, optional
        Determines whether processed images and masks should be returned by the function.

    write_dir: str, optional
        Path to directory where processed images and masks should be written. If not set, processed images and masks
        are returned by this function. Required if ``write_images=True``.

    image_export_format: {"dict", "native", "numpy"}, default: "dict"
        Return format for processed images and masks. ``"dict"`` returns dictionaries of images and masks as numpy
        arrays and associated characteristics. ``"native"`` returns images and masks in their internal format.
        ``"numpy"`` returns images and masks in numpy format. This argument is only used if ``export_images=True``.

    num_cpus: int, optional, default: None
        Number of CPU nodes that should be used for parallel processing. Image processing and feature computation can be
        parallelized using the ``ray`` package. If a ray cluster is defined by the user, this cluster will be used
        instead. By default, images are processed sequentially.

    **kwargs:
        Keyword arguments passed for importing images and masks (
        :func:`mirp.data_import.import_image_and_mask.import_image_and_mask`) and configuring settings:

        * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
        * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
        * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
        * image interpolation / resampling (
          :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
          :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
        * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
        * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)

    Returns
    -------
    None | list[Any]
        List of feature tables, if ``export_images=True``.

    See Also
    --------
    :func:`~mirp.extractFeaturesAndImages.extract_features_and_images`

    """
    return extract_features_and_images(
        write_features=False,
        export_features=False,
        write_images=write_images,
        export_images=export_images,
        write_dir=write_dir,
        image_export_format=image_export_format,
        num_cpus=num_cpus,
        **kwargs
    )


def extract_images_generator(
        write_images: bool = False,
        export_images: bool = True,
        write_dir: None | str = None,
        image_export_format: str = "dict",
        **kwargs
):
    """
    Process images and masks. This generator is a wrapper around
    :func:`~mirp.extractFeaturesAndImages.extract_features_and_images_generator`.

    Parameters
    ----------
    write_images: bool, default: True
       Determines whether processed images and masks should be written to the directory indicated by the
       ``write_dir`` keyword argument.

    export_images: bool, default: False
       Determines whether processed images and masks should be returned by the function.

    write_dir: str, optional
        Path to directory where processed images and masks should be written. If not set, processed images and masks
        are returned by this function. Required if ``write_images=True``.

    image_export_format: {"dict", "native", "numpy"}, default: "dict"
        Return format for processed images and masks. ``"dict"`` returns dictionaries of images and masks as numpy
        arrays and associated characteristics. ``"native"`` returns images and masks in their internal format.
        ``"numpy"`` returns images and masks in numpy format. This argument is only used if ``export_images=True``.

    **kwargs:
        Keyword arguments passed for importing images and masks (
        :func:`mirp.data_import.import_image_and_mask.import_image_and_mask`) and configuring settings:

        * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
        * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
        * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
        * image interpolation / resampling (
          :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
          :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
        * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
        * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)

    Yields
    ------
    None | list[Any]
       List of feature tables, if ``export_images=True``.

    See Also
    --------
    :func:`~mirp.extractFeaturesAndImages.extract_features_and_images_generator`

    """
    yield from extract_features_and_images_generator(
        write_features=False,
        export_features=False,
        write_images=write_images,
        export_images=export_images,
        write_dir=write_dir,
        image_export_format=image_export_format,
        **kwargs
    )


def extract_features_and_images(
        write_features: None | bool = None,
        export_features: None | bool = None,
        write_images: None | bool = None,
        export_images: None | bool = None,
        write_dir: None | str = None,
        image_export_format: str = "dict",
        num_cpus: None | int = None,
        parallel_backend: None | str = None,
        **kwargs
):
    """
    Processes images and computes features from regions of interest.

    Parameters
    ----------
    write_features: bool, optional
        Determines whether features computed from images should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    export_features: bool, optional
        Determines whether features computed from images should be returned by the function.

    write_images: bool, optional
        Determines whether processed images and masks should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    export_images: bool, optional
        Determines whether processed images and masks should be returned by the function.

    write_dir: str, optional
        Path to directory where features, processed images and masks should be written. If not set, features, processed images and masks
        are returned by this function. Required if ``write_features=True`` or ``write_images=True``.

    image_export_format: {"dict", "native", "numpy"}, default: "dict"
        Return format for processed images and masks. ``"dict"`` returns dictionaries of images and masks as numpy
        arrays and associated characteristics. ``"native"`` returns images and masks in their internal format.
        ``"numpy"`` returns images and masks in numpy format. This argument is only used if ``export_images=True``.

    num_cpus: int, optional, default: None
        Number of CPU nodes that should be used for parallel processing. Image and mask processing can be
        parallelized using the ``ray`` or ``joblib`` packages. If a ray cluster is defined by the user, this cluster
        will be used instead. By default, image and mask processing are processed sequentially.

    parallel_backend: {"none", "ray", "joblib"}, optional, default: "none"
        Type of backend to use. Default is the sequential backend (``"none"``). Alternative backends are ``"ray"`` and
        ``"joblib"``, which rely on the ray and joblib libraries respectively.

    **kwargs:
        Keyword arguments passed for importing images and masks (
        :func:`mirp.data_import.import_image_and_mask.import_image_and_mask`) and configuring settings:

        * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
        * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
        * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
        * image interpolation / resampling (
          :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
          :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
        * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
        * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)
        * feature computation / extraction (
          :class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`)

    Returns
    -------
    None | list[Any]
        List of features, images and masks, depending on ``export_features`` and ``export_images``.

    See Also
    --------
    Keyword arguments can be provided to configure the following:

    * image and mask import (:func:`~mirp.data_import.import_image_and_mask.import_image_and_mask`)
    * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
    * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
    * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
    * image interpolation / resampling (:class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
      :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
    * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
    * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)
    * feature computation / extraction (:class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`)

    """

    # Configure logger
    logging.basicConfig(
        format="%(levelname)s\t: %(processName)s \t %(asctime)s \t %(message)s",
        level=logging.INFO,
        stream=sys.stdout
    )

    backend = parse_parallel_backend(backend=parallel_backend, num_cpus=num_cpus)
    external_cluster = cluster_exists(backend=backend)
    start_parallel_cluster(backend=backend, num_cpus=num_cpus)

    # Switch to sequential backend if ray cluster is not formed somehow.
    if backend == "ray" and not cluster_exists(backend=backend):
        backend = "none"

    logging.info(message_parallel_process(backend=backend, num_cpus=num_cpus))

    if backend == "none":
        results = [
            workflow.standard_extraction(image_export_format=image_export_format)
            for workflow in _base_extract_features_and_images(
                write_features=write_features,
                export_features=export_features,
                write_images=write_images,
                export_images=export_images,
                write_dir=write_dir,
                **kwargs
            )
        ]

    elif backend == "ray":
        results = [
            _ray_extractor.remote(
                workflow=workflow,
                image_export_format=image_export_format
            )
            for workflow in _base_extract_features_and_images(
                write_features=write_features,
                export_features=export_features,
                write_images=write_images,
                export_images=export_images,
                write_dir=write_dir,
                **kwargs
            )
        ]
        results = ray_get(results)

    elif backend == "joblib":
        from joblib import Parallel, delayed
        results = Parallel(n_jobs=num_cpus)(
            delayed(workflow.standard_extraction)(
                image_export_format=image_export_format
            ) for workflow in _base_extract_features_and_images(
                write_features=write_features,
                export_features=export_features,
                write_images=write_images,
                export_images=export_images,
                write_dir=write_dir,
                **kwargs
            )
        )

    else:
        raise ValueError(f"parallel_backend is expected to be one of 'none', 'ray' or 'joblib'. Found: {backend}")

    if not external_cluster:
        shutdown_cluster(backend=backend)

    return results


def extract_features_and_images_generator(
        write_features: None | bool = None,
        export_features: None | bool = None,
        write_images: None | bool = None,
        export_images: None | bool = None,
        write_dir: None | str = None,
        image_export_format: str = "dict",
        num_cpus: None | int = None,
        parallel_backend: None | str = None,
        **kwargs
):
    """
    Processes images and computes features from regions of interest as a generator.

    Parameters
    ----------
    write_features: bool, optional
        Determines whether features computed from images should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    export_features: bool, optional
        Determines whether features computed from images should be returned by the function.

    write_images: bool, optional
        Determines whether processed images and masks should be written to the directory indicated by the
        ``write_dir`` keyword argument.

    export_images: bool, optional
        Determines whether processed images and masks should be returned by the function.

    write_dir: str, optional
        Path to directory where features, processed images and masks should be written. If not set, features, processed images and masks
        are returned by this function. Required if ``write_features=True`` or ``write_images=True``.

    image_export_format: {"dict", "native", "numpy"}, default: "dict"
        Return format for processed images and masks. ``"dict"`` returns dictionaries of images and masks as numpy
        arrays and associated characteristics. ``"native"`` returns images and masks in their internal format.
        ``"numpy"`` returns images and masks in numpy format. This argument is only used if ``export_images=True``.

    num_cpus: int, optional, default: None
        Number of CPU nodes that should be used for parallel processing. Image and mask processing can be
        parallelized using the ``joblib`` package. By default, image and mask processing are processed sequentially.

    parallel_backend: {"none", "joblib"}, optional, default: "none"
        Type of backend to use. Default is the sequential backend (``"none"``). ``"joblib"`` can be used as
        an alternative backend. ``"ray"`` cannot be used in a generator context, because only a single worker will be
        used.

    **kwargs:
        Keyword arguments passed for importing images and masks (
        :func:`mirp.data_import.import_image_and_mask.import_image_and_mask`) and configuring settings:

        * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
        * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
        * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
        * image interpolation / resampling (
          :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass` and
          :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
        * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
        * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)
        * feature computation / extraction (
          :class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`)

    Yields
    ------
    None | list[Any]
        List of features, images and masks, depending on ``export_features`` and ``export_images``.

    See Also
    --------
    Keyword arguments can be provided to configure the following:

    * image and mask import (:func:`~mirp.data_import.import_image_and_mask.import_image_and_mask`)
    * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
    * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
    * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
    * image interpolation / resampling (:class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass`
      and :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
    * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
    * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)
    * feature computation / extraction (:class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`)

    """

    # Configure logger
    logging.basicConfig(
        format="%(levelname)s\t: %(processName)s \t %(asctime)s \t %(message)s",
        level=logging.INFO,
        stream=sys.stdout
    )

    # Do not allow ray as a backend.
    backend = parse_parallel_backend(backend=parallel_backend, num_cpus=num_cpus, ray_allowed=False)
    external_cluster = cluster_exists(backend=backend)
    start_parallel_cluster(backend=backend, num_cpus=num_cpus)

    # Switch to sequential backend if ray cluster is not formed somehow.
    if backend == "ray" and not cluster_exists(backend=backend):
        backend = "none"

    logging.info(message_parallel_process(backend=backend, num_cpus=num_cpus))

    if backend == "none":
        for workflow in _base_extract_features_and_images(
            write_features=write_features,
            export_features=export_features,
            write_images=write_images,
            export_images=export_images,
            write_dir=write_dir,
            **kwargs
        ):
            yield workflow.standard_extraction(image_export_format=image_export_format)

    elif backend == "joblib":
        from joblib import Parallel, delayed
        parallel_gen = Parallel(n_jobs=num_cpus, return_as="generator")(
            delayed(workflow.standard_extraction)(
                image_export_format=image_export_format
            ) for workflow in _base_extract_features_and_images(
                write_features=write_features,
                export_features=export_features,
                write_images=write_images,
                export_images=export_images,
                write_dir=write_dir,
                **kwargs
            )
        )

        yield from parallel_gen

    else:
        raise ValueError(f"parallel_backend is expected to be one of 'none' or 'joblib'. Found: {backend}")

    if not external_cluster:
        shutdown_cluster(backend=backend)


@ray_remote
def _ray_extractor(workflow: StandardWorkflow, image_export_format="dict"):
    # Limit internal threading by third-party libraries.
    from mirp.utilities.parallel_ray import limit_inner_threads
    limit_inner_threads()

    return workflow.standard_extraction(image_export_format=image_export_format)


def _base_extract_features_and_images(
        image,
        mask=None,
        sample_name: None | str | list[str] = None,
        image_name: None | str | list[str] = None,
        image_file_type: None | str = None,
        image_modality: None | str | list[str] = None,
        image_sub_folder: None | str = None,
        mask_name: None | str | list[str] = None,
        mask_file_type: None | str = None,
        mask_modality: None | str | list[str] = None,
        mask_sub_folder: None | str = None,
        roi_name: None | str | list[str] | dict[str, str] = None,
        association_strategy: None | str | list[str] = None,
        settings: None | str | SettingsClass | list[SettingsClass] = None,
        stack_masks: str = "auto",
        stack_images: str = "auto",
        write_features: None | bool = None,
        export_features: None | bool = None,
        write_images: None | bool = None,
        export_images: None | bool = None,
        write_dir: None | str = None,
        **kwargs
):
    from mirp.data_import.import_image_and_mask import import_image_and_mask
    from mirp.settings.import_config_parameters import import_configuration_settings

    # Infer write_images, export_images, write_features, export_features based on write_dir.
    if write_images is None:
        write_images = write_dir is not None
    if export_images is None:
        export_images = write_dir is None
    if write_features is None:
        write_features = write_dir is not None
    if export_features is None:
        export_features = write_dir is None

    if not write_images and not write_features:
        write_dir = None

    if write_images and write_dir is None:
        raise ValueError("write_dir argument is required for writing images and masks, but not provided.")
    if write_features and write_dir is None:
        raise ValueError("write_dir argument is required for writing feature tables, but not provided.")

    if not write_features and not write_images and not export_features and not export_images:
        raise ValueError(
            f"At least one of write_features, write_images, export_features and export_images should be True."
        )

    # Import settings (to provide immediate feedback if something is amiss).
    if isinstance(settings, str):
        settings = import_configuration_settings(
            compute_features=write_features or export_features,
            path=settings
        )
    elif isinstance(settings, SettingsClass):
        settings = [settings]
    elif isinstance(settings, Iterable) and all(isinstance(x, SettingsClass) for x in settings):
        settings = list(settings)
    elif settings is None:
        settings = import_configuration_settings(
            compute_features=write_features or export_features,
            **kwargs
        )
    else:
        raise TypeError(
            f"The 'settings' argument is expected to be a path to a configuration xml file, a SettingsClass object, or "
            f"a list thereof. Found: {type(settings)}."
        )

    image_list = import_image_and_mask(
        image=image,
        mask=mask,
        sample_name=sample_name,
        image_name=image_name,
        image_file_type=image_file_type,
        image_modality=image_modality,
        image_sub_folder=image_sub_folder,
        mask_name=mask_name,
        mask_file_type=mask_file_type,
        mask_modality=mask_modality,
        mask_sub_folder=mask_sub_folder,
        roi_name=roi_name,
        association_strategy=association_strategy,
        stack_images=stack_images,
        stack_masks=stack_masks
    )

    yield from _generate_feature_and_image_extraction_workflows(
        image_list=image_list,
        settings=settings,
        write_dir=write_dir,
        write_features=write_features,
        export_features=export_features,
        write_images=write_images,
        export_images=export_images
    )


def _generate_feature_and_image_extraction_workflows(
        image_list: list[ImageFile],
        settings: list[SettingsClass],
        write_dir: None | str,
        write_features: bool,
        export_features: bool,
        write_images: bool,
        export_images: bool
) -> Generator[StandardWorkflow, None, None]:

    for image_file in image_list:

        # Check for potentially problematic stuff.
        image_file.check_associated_masks()

        for current_settings in settings:

            if not current_settings.feature_extr.has_any_feature_family() and (
                    current_settings.img_transform.spatial_filters is not None and not
                    current_settings.img_transform.feature_settings.has_any_feature_family()
            ) and (export_features or write_features):
                raise ValueError(
                    "No feature families were specified. Please set 'base_feature_families' or"
                    " 'response_map_feature_families'."
                )

            if current_settings.perturbation.noise_repetitions is None or \
                    current_settings.perturbation.noise_repetitions == 0:
                noise_repetition_ids = [None]
            else:
                noise_repetition_ids = list(range(current_settings.perturbation.noise_repetitions))

            if current_settings.perturbation.rotation_angles is None or len(
                    current_settings.perturbation.rotation_angles) == 0 or all(
                x == 0.0 for x in current_settings.perturbation.rotation_angles
            ):
                rotation_angles = [None]
            else:
                rotation_angles = copy.deepcopy(current_settings.perturbation.rotation_angles)

            if current_settings.perturbation.translation_fraction is None or len(
                current_settings.perturbation.translation_fraction) == 0 or all(
                x == 0.0 for x in current_settings.perturbation.translation_fraction
            ):
                translations = [None]
            else:
                config_translation = copy.deepcopy(current_settings.perturbation.translation_fraction)
                translations = []
                for translation_x in config_translation:
                    for translation_y in config_translation:
                        if not current_settings.general.by_slice:
                            for translation_z in config_translation:
                                translations += [(translation_z, translation_y, translation_x)]
                        else:
                            translations += [(0.0, translation_y, translation_x)]

            if current_settings.img_interpolate.new_spacing is None or len(
                    current_settings.img_interpolate.new_spacing) == 0 or all(
                x == 0.0 for x in current_settings.img_interpolate.new_spacing
            ):
                spacings = [None]
            else:
                spacings = copy.deepcopy(current_settings.img_interpolate.new_spacing)

            for noise_repetition_id in noise_repetition_ids:
                for rotation_angle in rotation_angles:
                    for translation in translations:
                        for spacing in spacings:
                            yield StandardWorkflow(
                                image_file=copy.deepcopy(image_file),
                                write_dir=write_dir,
                                settings=current_settings,
                                settings_name=current_settings.general.config_str,
                                write_features=write_features,
                                export_features=export_features,
                                write_images=write_images,
                                export_images=export_images,
                                noise_iteration_id=noise_repetition_id,
                                rotation=rotation_angle,
                                translation=translation,
                                new_image_spacing=spacing
                            )



import numpy as np


def rep(x, each=1, times=1, use_inversion=False):
    """"This functions replicates the R rep function for tiling and repeating vectors"""
    each = int(each)
    times = int(times)

    if each > 1:
        x = np.repeat(x, repeats=each)

    if times > 1:
        if use_inversion:
            len_x = len(x)
            y = np.zeros(len_x * times, dtype=x.dtype)
            ind_offset = 0
            for ii in np.arange(times):
                if ii % 2 == 0:
                    y[ind_offset:ind_offset+len_x] = x
                else:
                    y[ind_offset:ind_offset + len_x] = x[::-1]
                ind_offset += len_x
            x = y
        else:
            x = np.tile(x, reps=times)

    return x



import copy

import numpy as np
import pandas as pd
import skimage.measure

from mirp._features.texture_matrix import Matrix
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask


class MatrixSZM(Matrix):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Placeholders for derivative values computed using set_values_from_matrix
        # Size zone matrix
        self.sij: pd.DataFrame | None = None

        # Marginal sum over grey levels
        self.si: pd.DataFrame | None = None

        # Marginal sum over zone sizes
        self.sj: pd.DataFrame | None = None

        # Number of zones
        self.n_s: int | None = None

    def compute(
            self,
            image: GenericImage | None = None,
            mask: BaseMask | None = None,
            **kwargs
    ):
        from mirp.utilities.utilities import real_ndim

        # Define neighbour directions
        if self.spatial_method == "3d":
            connectivity = 3
            image = copy.deepcopy(image.get_voxel_grid())
            mask = copy.deepcopy(mask.roi_intensity.get_voxel_grid())

        elif self.spatial_method in ["2d", "2.5d"]:
            connectivity = 2
            image = image.get_voxel_grid()[self.slice_id, :, :]
            mask = mask.roi_intensity.get_voxel_grid()[self.slice_id, :, :]
        else:
            connectivity = -1  # Does nothing, because _spatial_method_error throws an error.
            self._spatial_method_error()

        # Check dimensionality and update connectivity if necessary.
        connectivity = min([connectivity, real_ndim(image)])

        # Set voxels outside roi to 0.0
        image[~mask] = 0.0

        # Count the number of voxels within the roi
        self.n_voxels = np.sum(mask)

        # Label all connected voxels with the same label.
        labelled_image = skimage.measure.label(
            label_image=image,
            background=0,
            connectivity=connectivity
        )

        # Generate data frame
        data = pd.DataFrame({
            "g": np.ravel(image),
            "label_id": np.ravel(labelled_image),
            "in_roi": np.ravel(mask)
        })

        # Remove all non-roi entries and count occurrence of combinations of volume id and grey level
        data = data[data.in_roi].groupby(by=["g", "label_id"]).size().reset_index(name="zone_size")

        # Count the number of co-occurring sizes and grey values
        matrix = data.groupby(by=["g", "zone_size"]).size().reset_index(name="n")

        # Rename columns
        matrix.columns = ["i", "j", "sij"]

        # Add matrix to object
        self.matrix = matrix

    def set_values_from_matrix(self, **kwargs):
        if self.is_empty():
            return

        # Copy of matrix
        self.sij = copy.deepcopy(self.matrix)

        # Sum over grey levels
        self.si = self.sij.groupby(by="i")["sij"].sum().reset_index().rename(columns={"sij": "si"})

        # Sum over zone sizes
        self.sj = self.sij.groupby(by="j")["sij"].sum().reset_index().rename(columns={"sij": "sj"})

        # Number of zones
        self.n_s = np.sum(self.sij.sij)  # Number of size zones

    @staticmethod
    def _get_grouping_columns():
        return ["i", "j"]



import numpy as np


def gaussian_preprocess_filter(
        orig_vox,
        orig_spacing,
        sample_spacing=None,
        param_beta=0.98,
        mode="nearest",
        separate_slices=False
):

    from scipy.ndimage import gaussian_filter

    # If no sample spacing is provided, assume original spacing. Note that for most purposes sample
    # spacing should be provided.
    if sample_spacing is None:
        sample_spacing = orig_spacing

    # Set sample spacing and orig_spacing to float
    sample_spacing = sample_spacing.astype(float)
    orig_spacing = orig_spacing.astype(float)

    # Calculate the zoom factors
    map_spacing = sample_spacing / orig_spacing

    # Only apply to down-sampling (map_spacing > 1.0)
    # map_spacing[map_spacing<=1.0] = 0.0

    # Don't filter along slices if calculations are to occur within the slice only
    if separate_slices:
        map_spacing[0] = 0.0

    # Calculate sigma
    sigma = np.sqrt(-8 * np.power(map_spacing, 2.0) * np.log(param_beta))

    # Apply filter
    new_vox = gaussian_filter(
        input=orig_vox.astype(np.float32),
        sigma=sigma,
        order=0,
        mode=mode
    )

    return new_vox



from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask


class Feature(object):

    def __init__(self, **kwargs):
        self.name: None | str = None
        self.abbr_name: None | str = None
        self.ibsi_id: None | str = None
        self.value: None | float = None
        self.table_name: None | str = None

        # Even though most features are IBSI-compliant, set value to False to avoid surprises in the future.
        self.ibsi_compliant: bool = False

    def clear_cache(self):
        pass

    def clear_local_cache(self, other):
        pass

    def compute(self, image: GenericImage, mask: BaseMask):
        raise NotImplementedError("compute method should be implemented in subclasses.")

    def create_table_name(self):
        self.table_name = "_".join(self._get_base_table_name_element())

    def update_ibsi_compliance(self):
        pass

    def is_ibsi_compliant(self, image: GenericImage) -> bool:
        # Update IBSI compliance based on attributes.
        self.update_ibsi_compliance()

        # Feature is compliant if it is compliant, and is derived from a image processed in an IBSI-compliant manner.
        return self.ibsi_compliant and image.ibsi_compliant

    def _get_base_table_name_element(self) -> list[str | None]:
        return [self.abbr_name]



from mirp._images.generic_image import GenericImage


class DXImage(GenericImage):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def update_separate_slices(self, x: bool):
        # Planar images should always separate slices.
        self.separate_slices = True


class CRImage(DXImage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class MGImage(DXImage):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import numpy as np

from mirp._data_import.generic_file import ImageFile
from mirp._data_import.utilities import flatten_list
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask


def read_image(
        image: ImageFile | GenericImage,
        to_numpy=False,
        **kwargs
) -> np.ndarray | GenericImage:
    image = image.to_object(**kwargs).promote()

    if to_numpy:
        image = image.get_voxel_grid()

    return image


def read_image_and_masks(
        image: ImageFile | GenericImage,
        to_numpy=False,
        **kwargs
) -> tuple[np.ndarray | GenericImage, list[np.ndarray] | list[BaseMask]]:
    # Read image from file.
    image_out = image.to_object(**kwargs).promote()

    mask_list = []
    if image.associated_masks is not None:
        mask_list = image.associated_masks

    # Read masks from file.
    if mask_list is not None:
        mask_list = [mask.to_object(image=image, **kwargs) for mask in mask_list]
        mask_list = flatten_list(mask_list)

    # Remove None entries.
    mask_list = [mask for mask in mask_list if mask is not None]

    if to_numpy:
        image_out = image_out.get_voxel_grid()
        mask_list = [mask.roi.get_voxel_grid() for mask in mask_list]

    return image_out, mask_list



import copy
import sys
if sys.version_info >= (3, 11):
    from typing import Unpack
else:
    from typing_extensions import Unpack

from mirp.settings.feature_parameters import FeatureExtractionSettingsClass
from mirp.settings.general_parameters import GeneralSettingsClass
from mirp.settings.image_processing_parameters import ImagePostProcessingClass
from mirp.settings.transformation_parameters import ImageTransformationSettingsClass
from mirp.settings.interpolation_parameters import ImageInterpolationSettingsClass, MaskInterpolationSettingsClass
from mirp.settings.resegmentation_parameters import ResegmentationSettingsClass
from mirp.settings.perturbation_parameters import ImagePerturbationSettingsClass


class SettingsClass:
    """
    Container for objects used to configure the image processing and feature processing workflow. This object can be
    initialised in two ways:

    * By providing (already initialised) configuration objects as arguments.
    * By passing arguments to configuration objects as keyword arguments. These configuration objects will then be
      created while initialising this container.

    Parameters
    ----------
    general_settings: GeneralSettingsClass, optional
        Configuration object for parameters related to the general process. See
        :class:`~mirp.settings.general_parameters.GeneralSettingsClass`.

    post_process_settings: ImagePostProcessingClass, optional
        Configuration object for parameters related to image (post-)processing. See
        :class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`.

    perturbation_settings: ImagePerturbationSettingsClass, optional
        Configuration object for parameters related to image perturbation / augmentation. See
        :class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`.

    img_interpolate_settings: ImageInterpolationSettingsClass, optional
        Configuration object for parameters related to image resampling. See
        :class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass`.

    roi_interpolate_settings: MaskInterpolationSettingsClass, optional
        Configuration object for parameters related to mask resampling. See
        :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`.

    roi_resegment_settings: ResegmentationSettingsClass, optional
        Configuration object for parameters related to mask resegmentation. See
        :class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`.

    feature_extr_settings: FeatureExtractionSettingsClass, optional
        Configuration object for parameters related to feature computation. See
        :class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`.

    img_transform_settings: ImageTransformationSettingsClass, optional
        Configuration object for parameters related to image transformation. See
        :class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`.

    **kwargs: Any, optional
        Keyword arguments for initialising configuration objects stored in this container object.

    See Also
    --------

    * general settings (:class:`~mirp.settings.general_parameters.GeneralSettingsClass`)
    * image post-processing (:class:`~mirp.settings.image_processing_parameters.ImagePostProcessingClass`)
    * image perturbation / augmentation (:class:`~mirp.settings.perturbation_parameters.ImagePerturbationSettingsClass`)
    * image interpolation / resampling (:class:`~mirp.settings.interpolation_parameters.ImageInterpolationSettingsClass`
      and :class:`~mirp.settings.interpolation_parameters.MaskInterpolationSettingsClass`)
    * mask resegmentation (:class:`~mirp.settings.resegmentation_parameters.ResegmentationSettingsClass`)
    * image transformation (:class:`~mirp.settings.transformation_parameters.ImageTransformationSettingsClass`)
    * feature computation / extraction (
      :class:`~mirp.settings.feature_parameters.FeatureExtractionSettingsClass`)

    """
    def __init__(
            self,
            general_settings: None | GeneralSettingsClass = None,
            post_process_settings: None | ImagePostProcessingClass = None,
            perturbation_settings: None | ImagePerturbationSettingsClass = None,
            img_interpolate_settings: None | ImageInterpolationSettingsClass = None,
            roi_interpolate_settings: None | MaskInterpolationSettingsClass = None,
            roi_resegment_settings: None | ResegmentationSettingsClass = None,
            feature_extr_settings: None | FeatureExtractionSettingsClass = None,
            img_transform_settings: None | ImageTransformationSettingsClass = None,
            **kwargs: Unpack[
                None | GeneralSettingsClass | ImagePostProcessingClass | ImagePostProcessingClass |
                ImageInterpolationSettingsClass | MaskInterpolationSettingsClass | ResegmentationSettingsClass |
                FeatureExtractionSettingsClass | ImageTransformationSettingsClass
            ]
    ):
        kwargs = copy.deepcopy(kwargs)

        # General settings.
        if general_settings is None:
            general_settings = GeneralSettingsClass(**kwargs)
        self.general = general_settings

        # Remove by_slice and no_approximation from the keyword arguments to avoid double passing.
        kwargs.pop("by_slice", None)
        kwargs.pop("no_approximation", None)
        kwargs.pop("ibsi_compliant", None)

        # Image interpolation settings.
        if img_interpolate_settings is None:
            img_interpolate_settings = ImageInterpolationSettingsClass(
                by_slice=general_settings.by_slice,
                **kwargs
            )
        self.img_interpolate = img_interpolate_settings

        # Mask interpolation settings.
        if roi_interpolate_settings is None:
            roi_interpolate_settings = MaskInterpolationSettingsClass(**kwargs)
        self.roi_interpolate = roi_interpolate_settings

        # Image (post-)processing settings.
        if post_process_settings is None:
            post_process_settings = ImagePostProcessingClass(**kwargs)
        self.post_process = post_process_settings

        # Image perturbation settings.
        if perturbation_settings is None:
            perturbation_settings = ImagePerturbationSettingsClass(**kwargs)
        self.perturbation = perturbation_settings

        # Mask resegmentation settings.
        if roi_resegment_settings is None:
            roi_resegment_settings = ResegmentationSettingsClass(**kwargs)
        self.roi_resegment = roi_resegment_settings

        # Feature extraction settings.
        if feature_extr_settings is None:
            feature_extr_settings = FeatureExtractionSettingsClass(
                by_slice=general_settings.by_slice,
                no_approximation=general_settings.no_approximation,
                ibsi_compliant=general_settings.ibsi_compliant,
                **kwargs
            )
        self.feature_extr = feature_extr_settings

        # Image transformation settings
        if img_transform_settings is None:
            img_transform_settings = ImageTransformationSettingsClass(
                by_slice=general_settings.by_slice,
                ibsi_compliant=general_settings.ibsi_compliant,
                **kwargs
            )
        self.img_transform = img_transform_settings

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False

        if self.general != other.general:
            return False
        if self.img_interpolate != other.img_interpolate:
            return False
        if self.roi_interpolate != other.roi_interpolate:
            return False
        if self.post_process != other.post_process:
            return False
        if self.perturbation != other.perturbation:
            return False
        if self.roi_resegment != other.roi_resegment:
            return False
        if self.feature_extr != other.feature_extr:
            return False
        if self.img_transform != other.img_transform:
            return False

        return True



from functools import lru_cache
from typing import Generator

import numpy as np

from mirp._features.szm_matrix import MatrixSZM
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.histogram import get_discretisation_parameters
from mirp._features.texture_features import FeatureTexture
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class FeatureSZM(FeatureTexture):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Perform close crop for SZM.
        self.cropping_distance = 0.0

    def get_matrix(
            self,
            image: GenericImage,
            mask: BaseMask,
    ) -> list[MatrixSZM]:
        # First discretise the image.
        image, mask = self.discretise_image(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_width=self.bin_width,
            bin_number=self.bin_number,
            cropping_distance=self.cropping_distance
        )

        # Then get matrix or matrices
        matrix_list = self._get_matrix(
            image=image,
            mask=mask,
            spatial_method=self.spatial_method
        )

        return matrix_list

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_matrix(
            image: GenericImage,
            mask: BaseMask,
            spatial_method: str
    ) -> list[MatrixSZM]:
        # Instantiate a helper copy of the current class to be able to use class methods without tying the cache to the
        # instance of the original object from which this method is called.
        matrix_instance = MatrixSZM(
            spatial_method=spatial_method
        )

        # Compute the required matrices.
        matrix_list = list(matrix_instance.generate(prototype=MatrixSZM, n_slices=image.image_dimension[0]))
        for matrix in matrix_list:
            matrix.compute(image=image, mask=mask)

        # Merge according to the spatial method.
        matrix_list = matrix_instance.merge(matrix_list, prototype=MatrixSZM)

        # Compute additional values from the individual matrices.
        for matrix in matrix_list:
            matrix.set_values_from_matrix()

        return matrix_list

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureSZM):
            self._get_matrix.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_matrix.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Compute or retrieve matrices from cache.
        matrices = self.get_matrix(image=image, mask=mask)

        # Compute feature value from matrices, and average over matrices.
        values = [self._compute(matrix=matrix) for matrix in matrices]
        if np.all(np.isnan(values)):
            self.value = np.nan
        else:
            self.value = np.nanmean(values)

    @staticmethod
    def _compute(matrix: MatrixSZM):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = (
                self._get_base_table_name_element()
                + self._get_spatial_table_name_element()
                + self._get_discretisation_table_name_element()
        )
        self.table_name = "_".join(table_elements)


class FeatureSZMSmallZoneEmphasis(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - small zone emphasis"
        self.abbr_name = "szm_sze"
        self.ibsi_id = "5QRC"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sj.sj / matrix.sj.j ** 2.0) / matrix.n_s


class FeatureSZMLargeZoneEmphasis(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - large zone emphasis"
        self.abbr_name = "szm_lze"
        self.ibsi_id = "48P8"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sj.sj * matrix.sj.j ** 2.0) / matrix.n_s


class FeatureSZMLowGreyLevelZoneEmphasis(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - low grey level zone emphasis"
        self.abbr_name = "szm_lgze"
        self.ibsi_id = "XMSY"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.si.si / matrix.si.i ** 2.0) / matrix.n_s


class FeatureSZMHighGreyLevelZoneEmphasis(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - high grey level zone emphasis "
        self.abbr_name = "szm_hgze"
        self.ibsi_id = "5GN9"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return  np.sum(matrix.si.si * matrix.si.i ** 2.0) / matrix.n_s


class FeatureSZMSmallZoneLowGreyLevelEmphasis(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - small zone low grey level emphasis"
        self.abbr_name = "szm_szlge"
        self.ibsi_id = "5RAI"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij / (matrix.sij.i * matrix.sij.j) ** 2.0) / matrix.n_s


class FeatureSZMSmallZoneHighGreyLevelEmphasis(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - small zone high grey level emphasis"
        self.abbr_name = "szm_szhge"
        self.ibsi_id = "HW1V"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij * matrix.sij.i ** 2.0 / matrix.sij.j ** 2.0) / matrix.n_s


class FeatureSZMLargeZoneLowGreyLevelEmphasis(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - large zone low grey level emphasis"
        self.abbr_name = "szm_lzlge"
        self.ibsi_id = "YH51"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij * matrix.sij.j ** 2.0 / matrix.sij.i ** 2.0) / matrix.n_s


class FeatureSZMLargeZoneHighGreyLevelEmphasis(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - large zone high grey level emphasis"
        self.abbr_name = "szm_lzhge"
        self.ibsi_id = "J17V"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij * matrix.sij.i ** 2.0 * matrix.sij.j ** 2.0) / matrix.n_s


class FeatureSZMGreyLevelNonUniformity(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - grey level non-uniformity"
        self.abbr_name = "szm_glnu"
        self.ibsi_id = "JNSA"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.si.si ** 2.0) / matrix.n_s


class FeatureSZMNormalisedGreyLevelNonUniformity(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - normalised grey level non-uniformity"
        self.abbr_name = "szm_glnu_norm"
        self.ibsi_id = "Y1RO"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.si.si ** 2.0) / matrix.n_s ** 2.0


class FeatureSZMZoneSizeNonUniformity(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - zone size non-uniformity"
        self.abbr_name = "szm_zsnu"
        self.ibsi_id = "4JP3"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sj.sj ** 2.0) / matrix.n_s


class FeatureSZMNormalisedZoneSizeNonUniformity(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - Normalised zone size non-uniformity"
        self.abbr_name = "szm_zsnu_norm"
        self.ibsi_id = "VB3A"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sj.sj ** 2.0) / matrix.n_s ** 2.0


class FeatureSZMZonePercentage(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - zone percentage"
        self.abbr_name = "szm_z_perc"
        self.ibsi_id = "P30P"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return matrix.n_s / matrix.n_voxels


class FeatureSZMGreyLevelVariance(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - grey level variance"
        self.abbr_name = "szm_gl_var"
        self.ibsi_id = "BYLV"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.sij.sij * matrix.sij.i) / matrix.n_s
        return np.sum((matrix.sij.i - mu) ** 2.0 * matrix.sij.sij) / matrix.n_s


class FeatureSZMZoneSizeVariance(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - zone size variance"
        self.abbr_name = "szm_zs_var"
        self.ibsi_id = "3NSA"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.sij.sij * matrix.sij.j) / matrix.n_s
        return np.sum((matrix.sij.j - mu) ** 2.0 * matrix.sij.sij) / matrix.n_s


class FeatureSZMZoneSizeEntropy(FeatureSZM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "SZM - zone size entropy"
        self.abbr_name = "szm_zs_entr"
        self.ibsi_id = "GU8N"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixSZM) -> float:
        if matrix.is_empty():
            return np.nan
        return -np.sum(matrix.sij.sij * np.log2(matrix.sij.sij / matrix.n_s)) / matrix.n_s


def get_szm_class_dict() -> dict[str, FeatureSZM]:
    class_dict = {
        "szm_sze": FeatureSZMSmallZoneEmphasis,
        "szm_lze": FeatureSZMLargeZoneEmphasis,
        "szm_lgze": FeatureSZMLowGreyLevelZoneEmphasis,
        "szm_hgze": FeatureSZMHighGreyLevelZoneEmphasis,
        "szm_szlge": FeatureSZMSmallZoneLowGreyLevelEmphasis,
        "szm_szhge": FeatureSZMSmallZoneHighGreyLevelEmphasis,
        "szm_lzlge": FeatureSZMLargeZoneLowGreyLevelEmphasis,
        "szm_lzhge": FeatureSZMLargeZoneHighGreyLevelEmphasis,
        "szm_glnu": FeatureSZMGreyLevelNonUniformity,
        "szm_glnu_norm": FeatureSZMNormalisedGreyLevelNonUniformity,
        "szm_zsnu": FeatureSZMZoneSizeNonUniformity,
        "szm_zsnu_norm": FeatureSZMNormalisedZoneSizeNonUniformity,
        "szm_z_perc": FeatureSZMZonePercentage,
        "szm_gl_var": FeatureSZMGreyLevelVariance,
        "szm_zs_var": FeatureSZMZoneSizeVariance,
        "szm_zs_entr": FeatureSZMZoneSizeEntropy
    }

    return class_dict


def generate_szm_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureSZM, None, None]:
    class_dict = get_szm_class_dict()
    szm_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_glszm_family():
        features = szm_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only SZM-features, and return if none are present.
    features = [feature for feature in features if feature in szm_features]
    if len(features) == 0:
        return

    # Features are parametrised by the choice of discretisation parameters and spatial methods.
    for discretisation_parameters in get_discretisation_parameters(
        settings=settings
    ):
        for spatial_method in settings.glszm_spatial_method:
            for feature in features:
                yield class_dict[feature](
                    spatial_method=spatial_method,
                    **discretisation_parameters
                )



from typing import Any
from pathlib import Path

import os
import pandas as pd

from mirp._data_import.generic_file import MaskFile


def extract_mask_labels(
        mask=None,
        sample_name: None | str | list[str] = None,
        mask_name: None | str | list[str] = None,
        mask_file_type: None | str = None,
        mask_modality: None | str | list[str] = None,
        mask_sub_folder: None | str = None,
        stack_masks: str = "auto",
        write_dir: None | str | Path = None
) -> pd.DataFrame | None:
    """
    Extract labels of regions of interest present in one or more mask files.

    Parameters
    ----------
    mask: Any
        A path to a mask file, a path to a directory containing mask files, a path to a config_data.xml
        file, a path to a csv file containing references to mask files, a pandas.DataFrame containing references to
        mask files, or a numpy.ndarray.

    sample_name: str or list of str, optional, default: None
        Name of expected sample names. This is used to select specific mask files. If None, no mask files are filtered
        based on the corresponding sample name (if known).

    mask_name: str, optional, default: None
        Pattern to match mask files against. The matches are exact. Use wildcard symbols ("*") to match varying
        structures. The sample name (if part of the file name) can also be specified using "#". For example,
        mask_name = '#_*_mask' would find John_Doe in John_Doe_CT_mask.nii or John_Doe_001_mask.nii. File extensions
        do not need to be specified. If None, file names are not used for filtering files and setting sample names.

    mask_file_type: {"dicom", "nifti", "nrrd", "numpy", "itk"}, optional, default: None
        The type of file that is expected. If None, the file type is not used for filtering files.
        "itk" comprises "nifti" and "nrrd" file types.

    mask_modality: {"rtstruct", "seg", "generic_mask"}, optional, default: None
        The type of modality that is expected. If None, modality is not used for filtering files.
        Note that only DICOM files contain metadata concerning modality. Masks from non-DICOM files are considered to
        be "generic_mask".

    mask_sub_folder: str, optional, default: None
        Fixed directory substructure where mask files are located. If None, the directory substructure is not used for
        filtering files.

    stack_masks: {"auto", "yes", "no"}, optional, default: "str"
        If mask files in the same directory cannot be assigned to different samples, and are 2D (slices) of the same
        size, they might belong to the same 3D mask stack. "auto" will stack 2D numpy arrays, but not other file
        types. "yes" will stack all files that contain 2D images, that have the same dimensions, orientation and
        spacing, except for DICOM files. "no" will not stack any files. DICOM files ignore this argument,
        because their stacking can be determined from metadata.

    write_dir: str, optional, default: None
        Directory to which a table with mask labels should be written, if any. Masks labels are exported to
        ``mask_labels.csv``.

    Returns
    -------
    pd.DataFrame | None
        The functions returns a table with labels extracted from mask files (``write_dir = None``) or nothing.

    """
    from mirp.data_import.import_mask import import_mask

    mask_list = import_mask(
        mask=mask,
        sample_name=sample_name,
        mask_name=mask_name,
        mask_file_type=mask_file_type,
        mask_modality=mask_modality,
        mask_sub_folder=mask_sub_folder,
        stack_masks=stack_masks
    )

    labels = [pd.DataFrame(_extract_mask_labels(ii, mask)) for ii, mask in enumerate(mask_list)]
    labels = pd.concat(labels)

    if write_dir is not None:
        write_dir = os.path.normpath(write_dir)
        if not os.path.exists(write_dir):
            os.makedirs(write_dir)

        labels.to_csv(
            path_or_buf=os.path.join(write_dir, "mask_labels.csv")
        )
    else:
        return labels


def _extract_mask_labels(index: int, mask: MaskFile) -> dict[str, Any]:

    labels = mask.export_roi_labels()
    labels.update({"mask_index": index})

    return labels



from typing import Any
from mirp._data_import.dicom_file import ImageDicomFile
from mirp._data_import.dicom_multi_frame import ImageDicomMultiFrame
from mirp._data_import.utilities import get_pydicom_meta_tag


class ImageDicomFileCT(ImageDicomFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def is_stackable(self, stack_images: str):
        return True

    def create(self):
        return self

    def load_data(self, **kwargs):
        self.image_data = self.load_data_generic()

    def export_metadata(self, self_only=False, **kwargs) -> None | dict[str, Any]:
        if not self_only:
            metadata = super().export_metadata()
        else:
            metadata = {}

        self.load_metadata()

        dcm_meta_data = []

        # Convolution kernel
        kernel = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1210),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9314)
        )
        if kernel is not None:
            dcm_meta_data += [("kernel", kernel)]

        # Reconstruction algorithm
        reconstruction_algorithm = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9315),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9314)
        )
        if reconstruction_algorithm is not None:
            dcm_meta_data += [("reconstruction_algorithm", reconstruction_algorithm)]

        # Peak kilo voltage output
        kvp = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0060),
            tag_type="float",
            macro_dcm_seq=(0x0018, 0x9325)
        )
        if kvp is not None:
            dcm_meta_data += [("kvp", kvp)]

        # Tube current in mA
        tube_current = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1151),
            tag_type="float"
        )
        if tube_current is not None:
            dcm_meta_data += [("tube_current", tube_current)]

        # Exposure time in milliseconds
        exposure_time = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1150),
            tag_type="float"
        )
        if exposure_time is not None:
            dcm_meta_data += [("exposure_time", exposure_time)]

        # Radiation exposure in mAs
        exposure = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1152),
            tag_type="float"
        )
        if exposure is not None:
            dcm_meta_data += [("exposure", exposure_time)]

        # Contrast/bolus agent
        contrast_agent = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x0010),
            tag_type="str"
        )
        if contrast_agent is not None:
            dcm_meta_data += [("contrast_agent", contrast_agent)]

        # Contrast/bolus agent phase
        contrast_agent_phase = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9344),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9341)
        )
        if contrast_agent_phase is not None:
            dcm_meta_data += [("contrast_agent_phase", contrast_agent_phase)]

        metadata.update(dict(dcm_meta_data))
        return metadata


class ImageDicomFileCTMultiFrame(ImageDicomMultiFrame, ImageDicomFileCT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import numpy as np
import copy

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import LawsTransformedImage
from mirp._imagefilters.generic import GenericFilter
from mirp._imagefilters.utilities import SeparableFilterSet, pool_voxel_grids
from mirp.settings.generic import SettingsClass


class LawsFilter(GenericFilter):
    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):

        super().__init__(image=image, settings=settings, name=name)

        self.ibsi_compliant = True
        self.ibsi_id = "JTXT"

        # Normalise kernel and energy filters? This is true by default (see IBSI).
        self.kernel_normalise = True
        self.energy_normalise = True

        # Set the filter name
        self.laws_kernel: None | str | list[str] = settings.img_transform.laws_kernel

        # Size of neighbourhood in chebyshev distance from center voxel
        self.delta: None | int | list[int] = settings.img_transform.laws_delta

        # Whether Laws texture energy should be calculated
        self.calculate_energy = settings.img_transform.laws_calculate_energy

        # Whether response maps or texture energy images should be made rotationally invariant
        self.rotation_invariance = settings.img_transform.laws_rotation_invariance

        # Which pooling method is used.
        self.pooling_method = settings.img_transform.laws_pooling_method

        # Set boundary condition
        self.mode = settings.img_transform.laws_boundary_condition

    def generate_object(self):
        # Generator for transformation objects.
        laws_kernel = copy.deepcopy(self.laws_kernel)
        if not isinstance(laws_kernel, list):
            laws_kernel = [laws_kernel]

        delta = copy.deepcopy(self.delta)
        if not isinstance(delta, list):
            delta = [delta]

        if not self.calculate_energy:
            delta = [None]

        # Iterate over options to yield filter objects with specific settings. A copy of the parent object is made to
        # avoid updating by reference.
        for current_laws_kernel in laws_kernel:
            for current_delta in delta:
                filter_object = copy.deepcopy(self)
                filter_object.laws_kernel = current_laws_kernel
                filter_object.delta = current_delta

                yield filter_object

    def transform(self, image: GenericImage) -> LawsTransformedImage:
        # Create placeholder Laws kernel response map.
        response_map = LawsTransformedImage(
            image_data=None,
            laws_kernel=self.laws_kernel,
            delta_parameter=self.delta,
            energy_map=self.calculate_energy,
            rotation_invariance=self.rotation_invariance,
            pooling_method=self.pooling_method,
            boundary_condition=self.mode,
            riesz_order=None,
            riesz_steering=None,
            riesz_sigma_parameter=None,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        # Initialise voxel grid.
        response_voxel_grid = None

        # Get filter list.
        filter_set_list: list[SeparableFilterSet] = self.get_filter_set().permute_filters(
            rotational_invariance=self.rotation_invariance)

        for ii, filter_set in enumerate(filter_set_list):
            # Convolve and compute response map.
            pooled_voxel_grid = filter_set.convolve(
                voxel_grid=image.get_voxel_grid(),
                mode=self.mode
            )

            # Pool grids.
            response_voxel_grid = pool_voxel_grids(
                x1=response_voxel_grid,
                x2=pooled_voxel_grid,
                pooling_method=self.pooling_method
            )

            # Remove img_laws_grid to explicitly release memory when collecting garbage.
            del pooled_voxel_grid

        if self.pooling_method == "mean":
            # Perform final pooling step for mean pooling.
            response_voxel_grid = np.divide(response_voxel_grid, len(filter_set_list))

        # Compute energy map from the response map.
        if self.calculate_energy:
            response_voxel_grid = self.response_to_energy(voxel_grid=response_voxel_grid)

        # Set voxel grid.
        response_map.set_voxel_grid(voxel_grid=response_voxel_grid)

        return response_map

    def response_to_energy(self, voxel_grid):

        # Take absolute value of the voxel grid.
        voxel_grid = np.abs(voxel_grid)

        # Set the filter size.
        filter_size = 2 * self.delta + 1

        # Set up the filter kernel.
        if self.energy_normalise:
            filter_kernel = np.ones(filter_size, dtype=float) / filter_size
        else:
            filter_kernel = np.ones(filter_size, dtype=float)

        # Create a filter set.
        if self.separate_slices:
            filter_set = SeparableFilterSet(
                filter_x=filter_kernel,
                filter_y=filter_kernel)
        else:
            filter_set = SeparableFilterSet(
                filter_x=filter_kernel,
                filter_y=filter_kernel,
                filter_z=filter_kernel)

        # Apply the filter.
        voxel_grid = filter_set.convolve(voxel_grid=voxel_grid, mode=self.mode)

        return voxel_grid

    def get_filter_set(self):

        # Get kernels
        kernels: str = copy.deepcopy(self.laws_kernel)

        # Deparse kernels to a list
        kernel_list = [kernels[ii:ii + 2] for ii in range(0, len(kernels), 2)]

        filter_x = None
        filter_y = None
        filter_z = None

        for ii, kernel in enumerate(kernel_list):
            if kernel.lower() == "l5":
                laws_kernel = np.array([1.0, 4.0, 6.0, 4.0, 1.0])
            elif kernel.lower() == "e5":
                laws_kernel = np.array([-1.0, -2.0, 0.0, 2.0, 1.0])
            elif kernel.lower() == "s5":
                laws_kernel = np.array([-1.0, 0.0, 2.0, 0.0, -1.0])
            elif kernel.lower() == "w5":
                laws_kernel = np.array([-1.0, 2.0, 0.0, -2.0, 1.0])
            elif kernel.lower() == "r5":
                laws_kernel = np.array([1.0, -4.0, 6.0, -4.0, 1.0])
            elif kernel.lower() == "l3":
                laws_kernel = np.array([1.0, 2.0, 1.0])
            elif kernel.lower() == "e3":
                laws_kernel = np.array([-1.0, 0.0, 1.0])
            elif kernel.lower() == "s3":
                laws_kernel = np.array([-1.0, 2.0, -1.0])
            else:
                raise ValueError(f"{kernel} is not an implemented Laws kernel")

            # Normalise kernel
            if self.kernel_normalise:
                laws_kernel /= np.sqrt(np.sum(np.power(laws_kernel, 2.0)))

            # Assign filter to variable.
            if ii == 0:
                filter_x = laws_kernel
            elif ii == 1:
                filter_y = laws_kernel
            elif ii == 2:
                filter_z = laws_kernel

        # Create FilterSet object
        return SeparableFilterSet(filter_x=filter_x,
                                  filter_y=filter_y,
                                  filter_z=filter_z)



import copy
import logging
import sys
import warnings

from typing import Generator, Any

import pandas as pd

from mirp.settings.generic import SettingsClass
from mirp._workflows.baseWorkflow import BaseWorkflow
from mirp._data_import.read_data import read_image_and_masks
from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import TransformedImage
from mirp._masks.base_mask import BaseMask
from mirp._image_processing.cropping import crop_image_to_size


class StandardWorkflow(BaseWorkflow):
    def __init__(
            self,
            settings: SettingsClass,
            settings_name: None | str = None,
            write_features: bool = False,
            export_features: bool = False,
            write_images: bool = False,
            export_images: bool = False,
            noise_iteration_id: None | int = None,
            rotation: None | float = None,
            translation: None | tuple[float, ...] = None,
            new_image_spacing: None | tuple[float, ...] = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.settings = settings
        self.settings_name = settings_name

        self.write_features = write_features
        self.export_features = export_features
        self.write_images = write_images
        self.export_images = export_images
        self.noise_iteration_id = noise_iteration_id
        self.rotation = rotation
        self.translation = translation
        self.new_image_spacing = new_image_spacing

    def _message_start(self):
        message_str = ["Initialising"]
        if (self.write_features or self.export_features) and (self.write_images or self.export_images):
            message_str += ["feature computation and image extraction"]
        elif self.write_features or self.export_features:
            message_str += ["feature computation"]
        elif self.write_images or self.export_images:
            message_str += ["image extraction"]
        else:
            raise ValueError("The workflow is not specified to do anything.")

        message_str += [f"using {self.image_file.modality} images"]

        if self.settings_name is not None and self.settings_name != "":
            message_str += [f"and configuration \"{self.settings_name}\""]

        message_str += [f"for {self.image_file.sample_name}."]

        return " ".join(message_str)

    def standard_image_processing(self) -> Generator[None | tuple[GenericImage, list[BaseMask]], None, None]:
        from mirp._image_processing.cropping import crop
        from mirp._image_processing.tissue_mask import create_tissue_mask
        from mirp._image_processing.alter_mask import alter_mask
        from mirp._image_processing.randomise_mask import randomise_mask
        from mirp._image_processing.split_mask import split_masks
        from mirp._data_import.utilities import flatten_list

        # Configure logger
        logging.basicConfig(
            format="%(levelname)s\t: %(processName)s \t %(asctime)s \t %(message)s",
            level=logging.INFO, stream=sys.stdout)

        # Notify
        logging.info(self._message_start())

        # Read image and masks.
        image, masks = read_image_and_masks(
            self.image_file,
            to_numpy=False,
            pet_suv_conversion=self.settings.post_process.suv_conversion_type
        )

        if masks is None or len(masks) == 0:
            warnings.warn("No segmentation masks were read.")
            return

        if any(not isinstance(mask, BaseMask) for mask in masks):
            object_types = list(set([str(type(mask)) for mask in masks if not isinstance(mask, BaseMask)]))
            raise TypeError(
                f"One or more mask objects are not of the expected BaseMask class. Found: {', '.join(object_types)}"
            )

        # Add type hints and remove masks that are empty.
        masks: list[BaseMask] = [mask for mask in masks if not mask.is_empty() and not mask.roi.is_empty_mask()]
        if len(masks) == 0:
            warnings.warn("No segmentation masks were read.")
            return

        # Set 2D or 3D processing.
        image.update_separate_slices(self.settings.general.by_slice)
        for mask in masks:
            mask.update_separate_slices(image.separate_slices)

        # Create a tissue mask
        if self.settings.post_process.bias_field_correction or \
                not self.settings.post_process.intensity_normalisation == "none":
            tissue_mask = create_tissue_mask(
                image=image,
                mask_type=self.settings.post_process.tissue_mask_type,
                mask_intensity_range=self.settings.post_process.tissue_mask_range
            )

            # Perform bias field correction
            if self.settings.post_process.bias_field_correction:
                image.bias_field_correction(
                    n_fitting_levels=self.settings.post_process.n_fitting_levels,
                    n_max_iterations=self.settings.post_process.n_max_iterations,
                    convergence_threshold=self.settings.post_process.convergence_threshold,
                    mask=tissue_mask,
                    in_place=True
                )

            image = image.normalise_intensities(
                normalisation_method=self.settings.post_process.intensity_normalisation,
                intensity_range=self.settings.post_process.intensity_normalisation_range,
                saturation_range=self.settings.post_process.intensity_normalisation_saturation,
                mask=tissue_mask
            )

        image = image.scale_intensities(
            scale=self.settings.post_process.intensity_scaling
        )

        # Estimate noise level.
        estimated_noise_level = self.settings.perturbation.noise_level
        if self.settings.perturbation.add_noise and estimated_noise_level is None:
            estimated_noise_level = image.estimate_noise()

        if self.settings.perturbation.add_noise:
            image.add_noise(noise_level=estimated_noise_level, noise_iteration_id=self.noise_iteration_id)

        # Translate, rotate and interpolate image
        image.interpolate(
            interpolate=self.settings.img_interpolate.interpolate,
            new_spacing=self.new_image_spacing if self.new_image_spacing is not None else image.image_spacing,
            translation=self.translation if self.translation is not None else (0.0, 0.0, 0.0),
            rotation=self.rotation if self.rotation is not None else 0.0,
            spline_order=self.settings.img_interpolate.spline_order,
            anti_aliasing=self.settings.img_interpolate.anti_aliasing,
            anti_aliasing_smoothing_beta=self.settings.img_interpolate.smoothing_beta
        )

        # Register masks to the interpolated images. Before registrations, masks may be oblique, and may not have a
        # one-to-one correspondence with the image space.
        [mask.register(
                image=image,
                spline_order=self.settings.roi_interpolate.spline_order,
                anti_aliasing=self.settings.img_interpolate.anti_aliasing,
                anti_aliasing_smoothing_beta=self.settings.img_interpolate.smoothing_beta
            ) for mask in masks]

        # Merge masks.
        if self.settings.general.mask_merge:
            masks = [masks[0].merge(masks)]

        # Split masks.
        if self.settings.general.mask_split:
            masks = flatten_list([mask.split_mask() for mask in masks])

        # Select largest region in each mask.
        if self.settings.general.mask_select_largest_region:
            [mask.select_largest_region() for mask in masks]

        # Select the axial slice with the largest portion of the ROI.
        if self.settings.general.mask_select_largest_slice:
            [mask.select_largest_slice() for mask in masks]

        # Crop slice stack
        if self.settings.perturbation.crop_around_roi:
            image, masks = crop(
                image=image,
                masks=masks,
                boundary=self.settings.perturbation.crop_distance
            )

        # Adapt roi sizes by dilation and erosion.
        masks = alter_mask(
            masks=masks,
            alteration_size=self.settings.perturbation.roi_adapt_size,
            alteration_method=self.settings.perturbation.roi_adapt_type,
            max_erosion=self.settings.perturbation.max_volume_erosion
        )

        # Update roi using SLIC
        if self.settings.perturbation.randomise_roi:
            masks = randomise_mask(
                image=image,
                masks=masks,
                repetitions=self.settings.perturbation.roi_random_rep
            )

        # Extract boundaries and tumour bulk
        masks = split_masks(
            masks=masks,
            boundary_sizes=self.settings.perturbation.roi_boundary_size,
            max_erosion=self.settings.perturbation.max_volume_erosion
        )

        # Resegmentise masks.
        [mask.resegmentise_mask(
            image=image,
            resegmentation_method=self.settings.roi_resegment.resegmentation_method,
            intensity_range=self.settings.roi_resegment.intensity_range,
            sigma=self.settings.roi_resegment.sigma
        ) for mask in masks]

        # self.extract_diagnostic_features(img_obj=img_obj, roi_list=roi_list, append_str="reseg")

        # Yield base image and masks
        yield image, masks

        # Create response maps
        if self.settings.img_transform.spatial_filters is not None:
            for transformed_image in self.transform_images(image=image):
                yield transformed_image, masks

    def transform_images(self, image: GenericImage) -> Generator[TransformedImage, None, None]:
        # Check if image transformation is required
        if self.settings.img_transform.spatial_filters is None:
            return

        # Get spatial filters to apply
        spatial_filter = self.settings.img_transform.spatial_filters

        # Iterate over spatial filters
        for current_filter in spatial_filter:

            if self.settings.img_transform.has_separable_wavelet_filter(x=current_filter):
                # Separable wavelet filters
                from mirp._imagefilters.separable_wavelet import SeparableWaveletFilter
                filter_obj = SeparableWaveletFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_nonseparable_wavelet_filter(x=current_filter):
                # Non-separable wavelet filters
                from mirp._imagefilters.nonseparable_wavelet import NonseparableWaveletFilter
                filter_obj = NonseparableWaveletFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_gaussian_filter(x=current_filter):
                # Gaussian filters
                from mirp._imagefilters.gaussian import GaussianFilter
                filter_obj = GaussianFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_laplacian_of_gaussian_filter(x=current_filter):
                # Laplacian of Gaussian filters
                from mirp._imagefilters.laplacian_of_gaussian import LaplacianOfGaussianFilter
                filter_obj = LaplacianOfGaussianFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_laws_filter(x=current_filter):
                # Laws' kernels
                from mirp._imagefilters.laws import LawsFilter
                filter_obj = LawsFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_gabor_filter(x=current_filter):
                # Gabor kernels
                from mirp._imagefilters.gabor import GaborFilter
                filter_obj = GaborFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_mean_filter(x=current_filter):
                # Mean / uniform filter
                from mirp._imagefilters.mean import MeanFilter
                filter_obj = MeanFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_square_transform_filter(x=current_filter):
                # Square transform filter
                from mirp._imagefilters.square_transform import SquareTransformFilter
                filter_obj = SquareTransformFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_square_root_transform_filter(x=current_filter):
                # Square root transform filter
                from mirp._imagefilters.square_root_transform import SquareRootTransformFilter
                filter_obj = SquareRootTransformFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_logarithm_transform_filter(x=current_filter):
                # Logarithm transform filter
                from mirp._imagefilters.log_transform import LogarithmTransformFilter
                filter_obj = LogarithmTransformFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_exponential_transform_filter(x=current_filter):
                # Exponential transform filter
                from mirp._imagefilters.exponential_transform import ExponentialTransformFilter
                filter_obj = ExponentialTransformFilter(image=image, settings=self.settings, name=current_filter)

            elif self.settings.img_transform.has_lbp_transform_filter(x=current_filter):
                # Local binary pattern filter
                from mirp._imagefilters.local_binary_patterns import LocalBinaryPatternFilter
                filter_obj = LocalBinaryPatternFilter(image=image, settings=self.settings, name=current_filter)

            else:
                raise ValueError(
                    f"{current_filter} is not implemented as a spatial filter. Please use one of ",
                    ", ".join(self.settings.img_transform.get_available_image_filters())
                )

            for current_filter_object in filter_obj.generate_object():
                # Create a response map.
                yield current_filter_object.transform(image=image)

    def standard_extraction(
            self,
            image_export_format: str = "dict",
            write_file_format: str = "nifti",
            write_all_masks: bool = False
    ):
        from mirp.utilities.utilities import random_string
        import os

        # Indicators to prevent the same masks from being written or exported multiple times.
        masks_written = False
        masks_exported = False

        # Placeholders
        feature_set_details = None
        feature_list: list[pd.DataFrame] = []
        image_list = []
        mask_list = []

        for image, masks in self.standard_image_processing():
            if image is None:
                continue

            # Type hinting
            image: GenericImage | TransformedImage = image
            masks: list[None | BaseMask] = masks

            if self.write_features or self.export_features:
                image_feature_list = []
                for mask in masks:
                    if mask is None:
                        continue

                    # Extract features, combine to a single DataFrame, and then add a roi_name for joining afterwards.
                    mask_feature_list = list(self._compute_radiomics_features(image=image, mask=mask))
                    if len(mask_feature_list) == 0:
                        continue
                    feature_set_details = self._get_feature_set_details(image=image, mask=mask)
                    mask_feature_set = pd.concat([feature_set_details] + mask_feature_list, axis=1)
                    image_feature_list += [mask_feature_set]

                if len(image_feature_list) > 0:
                    feature_list += [pd.concat(image_feature_list, axis=0, ignore_index=True)]

            if self.write_images:
                image.write(dir_path=self.write_dir, file_format=write_file_format)
                if not masks_written:
                    for mask in masks:
                        if mask is None:
                            continue
                        mask.write(dir_path=self.write_dir, file_format=write_file_format, write_all=write_all_masks)

                    # The standard_image_processing workflow only generates one set of masks - that which may change is
                    # image.
                    masks_written = True

            if self.export_images:
                image_list += [image.export(export_format=image_export_format)]
                if not masks_exported:
                    for mask in masks:
                        if mask is None:
                            continue
                        mask_list += [mask.export(write_all=write_all_masks, export_format=image_export_format)]
                        # The standard_image_processing workflow only generates one set of masks - that which may
                        # change is image. It is not necessary to export masks more than once.
                        masks_exported = True

        feature_set = None
        if (self.write_features or self.export_features) and len(feature_list) > 0:
            if len(feature_list) == 1:
                feature_set = feature_list[0]
            else:
                feature_set = feature_list[0]
                if feature_set_details is None:
                    raise ValueError("DEV: The feature_set_details variable has not been set.")

                shared_features = list(feature_set_details.columns)
                for ii in range(1, len(feature_list)):
                    feature_set = feature_set.merge(
                        feature_list[ii],
                        how="outer",
                        on=shared_features,
                        suffixes=(None, None)
                    )

        if self.write_features and isinstance(feature_set, pd.DataFrame):
            file_name = "_".join([image.sample_name, image.modality, random_string(k=16)]) + ".csv"
            # Check if the directory exists, and create otherwise.
            if not os.path.exists(self.write_dir):
                os.makedirs(self.write_dir)

            feature_set.to_csv(
                os.path.join(self.write_dir, file_name),
                sep=";",
                na_rep="",
                index=False
            )

        if self.export_features and self.export_images:
            return feature_set, image_list, mask_list
        elif self.export_features:
            return feature_set
        elif self.export_images:
            return image_list, mask_list

    def _compute_radiomics_features(
            self,
            image: GenericImage,
            mask: BaseMask
    ) -> Generator[pd.DataFrame, None, None]:
        from mirp._features.feature_generator import generate_features, feature_to_table

        # Check which feature settings to use.
        if isinstance(image, TransformedImage):
            feature_settings = self.settings.img_transform.feature_settings
        elif isinstance(image, GenericImage):
            feature_settings = self.settings.feature_extr
        else:
            raise TypeError(
                f"image is not a TransformedImage, GenericImage or a subclass thereof. Found: {type(image)}"
            )

        # Skip if no feature families are specified.
        if not feature_settings.has_any_feature_family():
            return

        # Get and compute features.
        features = list(generate_features(settings=feature_settings))

        # If IBSI-compliance is required, check that the feature is IBSI compliant and the input image was
        # processed in an IBSI-compliant manner. Filter those features where this is not the case.
        if feature_settings.ibsi_compliant:
            features = [
                feature for feature in features
                if feature.is_ibsi_compliant(image=image) or not feature_settings.ibsi_compliant
            ]

        if len(features) == 0:
            return

        previous_feature = None
        for feature in features:
            feature.compute(image=image, mask=mask)
            if previous_feature is not None:
                previous_feature.clear_local_cache(other=feature)
            previous_feature = feature

        # Make a final pass to empty cache to prevent memory leaks.
        for feature in features:
            feature.clear_cache()

        # Convert to table and add image descriptors.
        yield image.parse_feature_names(feature_to_table(features))

    def _get_feature_set_details(
            self,
            image: GenericImage,
            mask: BaseMask
    ) -> pd.DataFrame:

        def _from_dict(
                x: dict[str, str],
                key: str,
                default: Any = None,
                ii: None | int = None
        ):
            value = x[key] if key in x else default
            if value is not None and ii is not None:
                value = value[ii]

            if value is not None and isinstance(value, str) and value == "":
                value = default

            return value

        image_metadata = image.get_export_attributes()
        mask_metadata = mask.get_export_attributes()

        # The name of the current settings configuration derives from the workflow, not image or mask.
        settings_name = _from_dict({"settings_name": self.settings_name}, "settings_name")

        feature_set_details = {
            "sample_name": _from_dict(image_metadata, "sample_name"),
            "image_file_name": _from_dict(image_metadata, "file_name"),
            "image_directory": _from_dict(image_metadata, "dir_path"),
            "image_study_date": _from_dict(image_metadata, "study_date"),
            "image_study_description": _from_dict(image_metadata, "study_description"),
            "image_series_description": _from_dict(image_metadata, "series_description"),
            "image_series_instance_uid": _from_dict(image_metadata, "series_instance_uid"),
            "image_modality": _from_dict(image_metadata, "modality", "generic"),
            "image_pet_suv_type": _from_dict(image_metadata, "suv_type"),
            "image_mask_label": _from_dict(mask_metadata, "roi_name"),
            "image_mask_file_name": _from_dict(mask_metadata, "file_name"),
            "image_mask_directory": _from_dict(mask_metadata, "dir_path"),
            "image_mask_series_description": _from_dict(mask_metadata, "series_description"),
            "image_mask_series_instance_uid": _from_dict(mask_metadata, "series_instance_uid"),
            "image_settings_id": settings_name,
            "image_voxel_size_x": _from_dict(image_metadata, "image_spacing", ii=2),
            "image_voxel_size_y": _from_dict(image_metadata, "image_spacing", ii=1),
            "image_voxel_size_z": _from_dict(image_metadata, "image_spacing", ii=0),
            "image_noise_level": _from_dict(image_metadata, "noise_level"),
            "image_noise_iteration_id": _from_dict(image_metadata, "noise_id"),
            "image_rotation_angle": _from_dict(image_metadata, "rotation", default=0.0),
            "image_translation_x": _from_dict(image_metadata, "translation", ii=2, default=0.0),
            "image_translation_y": _from_dict(image_metadata, "translation", ii=1, default=0.0),
            "image_translation_z": _from_dict(image_metadata, "translation", ii=0, default=0.0),
            "image_mask_randomise_id": _from_dict(mask_metadata, "mask_randomisation_id"),
            "image_mask_adapt_size": _from_dict(mask_metadata, "mask_alteration_size", default=0.0)
        }

        return pd.DataFrame(feature_set_details, index=[0])

    def deep_learning_conversion(
        self,
        output_slices: bool = False,
        crop_size: None | list[float] | list[int] = None,
        image_export_format: str = "numpy",
        write_file_format: str = "numpy",
    ):
        # Placeholders
        image_list = []
        mask_list = []

        for image, mask in self._deep_learning_conversion(
            output_slices=output_slices,
            crop_size=crop_size
        ):
            if image is None:
                continue
            if mask is None:
                continue

            if self.write_images:
                image.write(dir_path=self.write_dir, file_format=write_file_format)
                mask.write(dir_path=self.write_dir, file_format=write_file_format, write_all=False)

            if self.export_images:
                image_list += [image.export(export_format=image_export_format)]
                mask_list += [mask.export(write_all=False, export_format=image_export_format)]

        if self.export_images:
            return image_list, mask_list

    def _deep_learning_conversion(
            self,
            output_slices: bool = False,
            crop_size: None | list[float] | list[int] = None
    ) -> Generator[tuple[GenericImage, BaseMask], None, None]:
        from mirp._image_processing.cropping import crop

        # Avoid updating by reference.
        crop_size = copy.deepcopy(crop_size)
        pre_crop_size = []

        # Set crop_size.
        if crop_size is None and output_slices:
            crop_size = [None, None]
            remove_empty_slices = False

        elif crop_size is None and not output_slices:
            crop_size = [None, None, None]
            remove_empty_slices = False

        elif len(crop_size) == 1 and output_slices:
            crop_size = [crop_size[0], crop_size[0]]
            remove_empty_slices = True

        elif len(crop_size) == 1 and not output_slices:
            crop_size = [crop_size[0], crop_size[0], crop_size[0]]
            remove_empty_slices = False

        elif len(crop_size) == 2 and output_slices:
            crop_size = [crop_size[0], crop_size[1]]
            remove_empty_slices = True

        elif len(crop_size) == 2 and not output_slices:
            crop_size = [None, crop_size[0], crop_size[1]]
            remove_empty_slices = True

        elif len(crop_size) == 3 and output_slices:
            pre_crop_size = [crop_size[0], None, None]
            crop_size = [crop_size[1], crop_size[2]]
            remove_empty_slices = False

        elif len(crop_size) == 3 and not output_slices:
            crop_size = [crop_size[0], crop_size[1], crop_size[2]]
            remove_empty_slices = False
        else:
            raise ValueError(f"The crop_size argument is longer than 3: {len(crop_size)}")

        for image, masks in self.standard_image_processing():
            if image is None:
                continue
            for mask in masks:
                if mask is None:
                    continue

                # Type hinting
                image: GenericImage = image
                mask: BaseMask = mask

                # Pre-crop the image
                if len(pre_crop_size) > 0:
                    image, mask = crop_image_to_size(
                        image=image,
                        masks=mask,
                        crop_size=pre_crop_size,
                        crop_center=mask.get_center_position()
                    )

                # Remove slices where the mask is not present.
                if remove_empty_slices:
                    image, mask = crop(
                        image=image,
                        masks=mask,
                        boundary=0.0,
                        z_only=True
                    )

                # Find the overall crop center.
                crop_center = mask.get_center_position()

                if len(crop_size) == 2:
                    # 2D-cropping
                    image_slices = image.get_slices()
                    mask_slices = mask.get_slices(primary_mask_only=True)

                    # Update crop center.
                    crop_center = [crop_center[1], crop_center[2]]

                    if image_slices is None or len(image_slices) == 0 or mask_slices is None or len(mask_slices) == 0:
                        continue

                    for ii, image_slice in enumerate(image_slices):
                        yield crop_image_to_size(
                            image=image_slice,
                            masks=mask_slices[ii],
                            crop_size=crop_size,
                            crop_center=crop_center
                        )
                else:
                    # 3D-cropping.
                    yield crop_image_to_size(
                        image=image,
                        masks=mask,
                        crop_size=crop_size,
                        crop_center=crop_center
                    )
